
/* mkmkn.c v0.9b (c)  .., , 2019-2020            */
/*  161eForth     eforth.mkn  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define REGIONS (30)
#define MAXMKN (524292)
#define WORDS (212)

FILE *fIn, *fOut;
static unsigned char line[512];             /*     */
static unsigned char text[65536];           /*    */

static unsigned char mknImage[MAXMKN];      /*  eforth.mkn */

int tregs[1000], lregs[1000], rregs[1000], ptext;    /*  R9000..R9999 */
int fregions[REGIONS], lregions[REGIONS];            /*       */

int iRegion;                                /*    */


int pTokens[WORDS], lTokens[WORDS];         /*           */
int iToken;                                 /*    */


void Init(void)
{ int i;
  for (i=0; i<1000; i++)                    /*    */
  { tregs[i]=0;
    lregs[i]=0;
    rregs[i]=0;
  }
  for (i=0; i<REGIONS; i++)
  { fregions[i]=0;                          /*   */
    lregions[i]=0;
  }
  iRegion = 0;                              /*    1      */
  iToken = 0;                               /*    1       */
  ptext = 0;                                /*   text[0]        */
  for (i=0; i<MAXMKN; i++)                  /*    */
    mknImage[i]=0;
}


void Pack2Ch(unsigned char *s, int i1, int i2) /*  2   3  */
{ s[0] = (i1/100)*16 + (i1/10)%10;
  s[1] = (i1%10)*16 + (i2/100);
  s[2] = ((i2/10)%10)*16 + (i2%10);
}


void PackDec(unsigned char *src, unsigned char *dest) /*  4    */
{ dest[0]=0;  /*  */
  dest[7]=11; /*        */
  Pack2Ch(dest+1, src[0], src[1]);
  Pack2Ch(dest+4, src[2], src[3]);
}


void Normalize(unsigned char *s) /*    */
{ if ((s[1]!=0) || (s[2]!=0) || (s[3]!=0) ||
      (s[4]!=0) || (s[5]!=0) || (s[6]!=0)) /* if  0 */
    while (s[1]<16)
    { s[1]=(((s[1]&15)<<4) | (s[2]>>4));
      s[2]=(((s[2]&15)<<4) | (s[3]>>4));
      s[3]=(((s[3]&15)<<4) | (s[4]>>4));
      s[4]=(((s[4]&15)<<4) | (s[5]>>4));
      s[5]=(((s[5]&15)<<4) | (s[6]>>4));
      s[6]= ((s[6]&15)<<4);
      (s[7])--;
    }
  else s[7]=0; /*  s==0,   */
}


int PackStr(unsigned char *src, unsigned char **d)
/*   str   ,   d */
{ static unsigned char str[4];
  int n=0, i=0;
  str[0]=0;  str[1]=0;  str[2]=0;  str[3]=0;
  while(*src)
  { if ((src[0]=='\\') && (src[1]=='n')) /* \n    */
    { str[i++] = 13;  n++; /*  CR */
      if (i==4)
      { PackDec(str,*d);  Normalize(*d);  *d += 8;
        str[0]=0;  str[1]=0;  str[2]=0;  str[3]=0;
        i = 0;
      }
      src++; /*    */
      str[i++] = 10;  n++; /*  LF */
      if (i==4)
      { PackDec(str,*d);  Normalize(*d);  *d += 8;
        str[0]=0;  str[1]=0;  str[2]=0;  str[3]=0;
        i = 0;
      }
    } else if ((*src!=13) && (*src!=10)) /*     */
    { str[i++] = *src;  n++;
      if (i==4)
      { PackDec(str,*d);  Normalize(*d);  *d += 8;
        str[0]=0;  str[1]=0;  str[2]=0;  str[3]=0;
        i = 0;
      }
    }
    src++;
  }
  if (i)
  { PackDec(str,*d);  Normalize(*d);  *d += 8;
  }
  return n; /*    */
}


int IsBlank(unsigned char *s)
{ int i=0;
  unsigned char c=s[0];
  while (c)
  { if ((c!=32) && (c!=13) && (c!=10) && (c!=9)) return 0; /*  s   */
    c=s[i++];
  }
  return 1;                                /*  s  */
}


void StoreRegion(unsigned char *s)          /*     */
{ unsigned char *p = text + ptext;
  fregions[++iRegion] = ptext;
  lregions[iRegion] = PackStr(s, &p);
  ptext = p - text;
}

void StoreToken(unsigned char *s)           /*    */
{ unsigned char *p = text + ptext;
  pTokens[++iToken] = ptext;
  lTokens[iToken] = PackStr(s, &p);
  ptext = p - text;
}

void StoreRegister(unsigned char *s)        /*    */
{ if (s[0]!='R') fprintf(stderr, "Wrong register: %s|", line);
  else
  { int n = (((s[1]-'0')*10+s[2]-'0')*10+s[3]-'0')*10+s[4]-'0'; /*   */
    unsigned char *p = text + ptext;
    rregs[n-9000] = iRegion;
    tregs[n-9000] = ptext;
    lregs[n-9000] = PackStr(s+7, &p);
    ptext = p - text;
  }
}


void MkReg(int reg, int nn, int aaaaa, int lllll)
/* Rn := NNAAAAALLLLL */
{ unsigned char *p = mknImage+4+reg*8;
  p[0] = 0;  /*  */
  p[7] = 11; /*    */
  p[1] = ((nn/10)<<4) + (nn%10);
  p[2] = ((aaaaa/10000)<<4) + ((aaaaa/1000)%10);
  p[3] = (((aaaaa/100)%10)<<4) + ((aaaaa/10)%10);
  p[4] = ((aaaaa%10)<<4) + (lllll/10000);
  p[5] = (((lllll/1000)%10)<<4) + ((lllll/100)%10);
  p[6] = (((lllll/10)%10)<<4) + (lllll%10);
  Normalize(p); /*    */
}


void MkMKN(void)
/*        */
{ int i;
  int last = 251 + (iRegion + iToken + ptext/8)/4; /* 1  ,   MK.EXE */
  mknImage[0] = 0;                         /*    0 */
  mknImage[1] = 0;
  mknImage[2] = (last&255);                /*   */
  mknImage[3] = (last>>8);

  for (i=0; i<1000; i++)                   /*    */
    if (lregs[i]) /*       */
      MkReg(i, rregs[i], 1001+iRegion+iToken+tregs[i]/8, lregs[i]);

  MkReg(1000, 0, 0, 1001+iRegion);            /*    */

  for (i=1; i<=iRegion; i++)                  /*    */
    MkReg(1000+i, 0, 1001+iRegion+iToken+fregions[i]/8, lregions[i]);

  for (i=1; i<=iToken; i++)                    /*    */
    MkReg(1000+iRegion+i, 0, 1001+iRegion+iToken+pTokens[i]/8, lTokens[i]);

  fwrite(mknImage, 1, 4, fOut);               /*   */
  fwrite(mknImage+4, 8, 1001+iRegion+iToken, fOut);  /*   */
  fwrite(text, 8, ptext/8, fOut);             /*   */

  i = (1 + iRegion + iToken + ptext/8)%4;
  fwrite(mknImage+4, 8, 8-i, fOut); /* align last group, one extra group */
}


int main(int arc, char *argv[])
{ fIn = fopen(argv[1],"rb");
  fprintf(stderr, "mkmkn v0.8b (c) Ilya V. Vasilyev\n");
  Init();
  if (fIn!=NULL)
  { while (fgets((char *)line, 512, fIn)!=NULL) /*    */
      if (IsBlank(line)==0)
      { StoreRegion(line);                /*    --   */
        while (fgets((char *)line, 512, fIn)!=NULL)
        { if (IsBlank(line)) break;       /*      */
          StoreRegister(line);
        }
      }
    fclose(fIn);
    fprintf(stderr, "text = %d\n", ptext/8);

    fIn=fopen(argv[2],"rb");
    if (fIn!=NULL)
    { while (fgets((char *)line, 512, fIn)!=NULL) /*    */
        StoreToken(line);
      fclose(fIn);
      fprintf(stderr, "tokens = %d/%d; text = %d\n", iToken, WORDS, ptext/8);

      fOut = fopen(argv[3], "wb");
      if (fOut!=NULL)
      { MkMKN();
        fclose(fOut);
        return 1;
      } else fprintf(stderr, "Can't create file: %s\n", argv[3]);
    } else fprintf(stderr, "File not found: %s\n", argv[2]);
  } else fprintf(stderr, "File not found: %s\n", argv[1]);
  fprintf(stderr, "Usage: mkmkn regs.txt tokens.txt outfile.mkn\n");
  return 0;
}
